unit ExcelLib;

interface

uses
  Classes, Graphics, Generics.Collections, Generics.Defaults;

  // ó : http://www.codeproject.com/KB/office/biffcsharp.aspx

const
  COLOR_DEFAULT = $7fff;
  COLOR_BLACK = $0;
  COLOR_WHITE = $1;
  COLOR_RED = $2;
  COLOR_GREEN = $3;
  COLOR_BLUE = $4;
  COLOR_YELLOW = $5;
  COLOR_MAGENTA = $6;
  COLOR_CYAN = $7;
  COLOR_DARKRED = $10;
  COLOR_DARKGREEN = $11;
  COLOR_DARKBLUE = $12;
  COLOR_OLIVE = $13;
  COLOR_PURPLE = $14;
  COLOR_TEAL = $15;
  COLOR_SILVER = $16;
  COLOR_GRAY = $17;
  COLOR_WINDOWTEXT = $18;
  COLOR_WINDOWBACKGROUND = $19;

  RECORD_BOF = $0209;
  RECORD_EOF = $0A;
  RECORD_CODEPAGE = $0042;
  RECORD_FONT = $0231;
  RECORD_HEADER = $0014;
  RECORD_FOOTER = $0015;
  RECORD_FORMAT = $001E;
  RECORD_WINDOWPROTECT = $0019;
  RECORD_EXTENDED = $0243;
  RECORD_STYLE = $0293;
  RECORD_COMUMNINFO = $007D;

  RECORD_LABEL = $0204;
  RECORD_EMPTY = $0201;
  RECORD_NUMBER = $0203;

  clBegin: array [0..5] of Word = (RECORD_BOF, 8, 0, $10, 0, 0);
  clEnd: array [0..1] of Word = (RECORD_EOF, 0);

type
  TExcelDocument = class;
  TFontInfo = class;
  TAlignment = ( xGeneral, xLeft, xCentered, xRight, xFilled );

  TCellInfo = class
  private
    FForeColor: Word;
    FBackColor: Word;
    FDocument: TExcelDocument;
    FFXIndex: Byte;
    FFont: TFontInfo;
    FFormat: string;
    FAlignment: TAlignment;
    FValue: Variant;
    FRow: Integer;
    FColumn: Integer;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetBackColor(const Value: Word);
    procedure SetColumn(const Value: Integer);
    procedure SetDocument(const Value: TExcelDocument);
    procedure SetFont(const Value: TFontInfo);
    procedure SetForeColor(const Value: Word);
    procedure SetFormat(const Value: string);
    procedure SetFXIndex(const Value: Byte);
    procedure SetRow(const Value: Integer);
    procedure SetValue(const Value: Variant);
  public
    constructor Create( AExcelDocument: TExcelDocument );
    destructor Destroy; override;

    property FXIndex: Byte read FFXIndex write SetFXIndex;
    property Document: TExcelDocument read FDocument write SetDocument;
    property Value: Variant read FValue write SetValue;
    property Format: string read FFormat write SetFormat;
    property BackColor: Word read FBackColor write SetBackColor;
    property ForeColor: Word read FForeColor write SetForeColor;
    property Row: Integer read FRow write SetRow;
    property Column: Integer read FColumn write SetColumn;
    property Font: TFontInfo read FFont write SetFont;
    property Alignment: TAlignment read FAlignment write SetAlignment;
  end;

  TCell = class
  private
    FDocument: TExcelDocument;
    FCellInfo: TCellInfo;
    function GetAlignment: TAlignment;
    function GetBackColor: Word;
    function GetColumn: Integer;
    function GetFont: TFontInfo;
    function GetForeColor: Word;
    function GetFormat: string;
    function GetRow: Integer;
    function GetValue: Variant;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetBackColor(const Value: Word);
    procedure SetFont(const Value: TFontInfo);
    procedure SetForeColor(const Value: Word);
    procedure SetFormat(const Value: string);
    procedure SetValue(const Value: Variant);
  public
    constructor Create( const ARow, AColumn: Integer; ADocument: TExcelDocument );

    property Value: Variant read GetValue write SetValue;
    property Format: string read GetFormat write SetFormat;
    property BackColor: Word read GetBackColor write SetBackColor;
    property ForeColor: Word read GetForeColor write SetForeColor;
    property Font: TFontInfo read GetFont write SetFont;
    property Alignment: TAlignment read GetAlignment write SetAlignment;
    property Row: Integer read GetRow;
    property Column: Integer read GetColumn;
  end;

  TColumnInfo = class
  private
    FColumnIndex: Integer;
    FWidth: Integer;
    procedure SetColumnIndex(const Value: Integer);
    procedure SetWidth(const Value: Integer);
  public
    function Equals(Obj: TObject): Boolean; override;

    property ColumnIndex: Integer read FColumnIndex write SetColumnIndex;
    property Width: Integer read FWidth write SetWidth;
  end;
  TColumnInfoComparer = TComparer<TColumnInfo>;

  TFontInfo = class
  private
    FFontSize: Integer;
    FFontName: String;
    FFontColor: Word;
    FFontStyle: TFontStyles;
    procedure SetFontName(const Value: String);
    procedure SetFontSize(const Value: Integer);
    procedure SetFontColor(const Value: Word);
    procedure SetFontStyle(const Value: TFontStyles);
  public
    constructor Create; overload;
    constructor Create( const AFontName: string; const AFontSize: Integer; const AFontColor: Word ); overload;

    procedure Assign( AFontInfo: TFontInfo );
    function Equals(Obj: TObject): Boolean; override;

    property FontName: String read FFontName write SetFontName;
    property FontSize: Integer read FFontSize write SetFontSize;
    property FontStyle: TFontStyles read FFontStyle write SetFontStyle;
    property FontColor: Word read FFontColor write SetFontColor;
  end;
  TFontInfoComparer = TComparer<TFontInfo>;

  TFormatInfo = class
  private
    FCellInfo: TCellInfo;
    FFormatIndex: Integer;
    FFontIndex: Integer;
    function GetAlignment: TAlignment;
    function GetBackColor: Word;
    function GetFont: TFontInfo;
    function GetForeColor: Word;
    function GetFormat: string;
  public
    constructor Create( ACellInfo: TCellInfo );

    function Equals(Obj: TObject): Boolean; override;

    property BackColor: Word read GetBackColor;
    property ForeColor: Word read GetForeColor;
    property Font: TFontInfo read GetFont;
    property Format: string read GetFormat;
    property Alignment: TAlignment read GetAlignment;
    property FormatIndex: Integer read FFormatIndex;
    property FontIndex: Integer read FFontIndex;
  end;
  TFormatInfoComparer = TComparer<TFormatInfo>;

  TExcelDocument = class
  private
    FCells: TDictionary<Int64, TCellInfo>;
    FFormats: TList<string>;
    FFonts: TObjectList<TFontInfo>;
    FFxs: TObjectList<TFormatInfo>;
    FRows: TList<Integer>;
    FColumns: TObjectList<TColumnInfo>;
    FCodePage: Word;
    FUserName: string;

    procedure SetCodePage(const Value: Word);
    procedure SetUserName(const Value: string);
    procedure BuildInternalTables;

    procedure WriteWordArray( AStream: TStream; const AArray: array of Word );
    procedure WriteByteArray( AStream: TStream; const AArray: array of Byte );

    procedure WriteAuthorRecord( AStream: TStream );
    procedure WriteCodepageRecord( AStream: TStream );
    procedure WriteFontTable( AStream: TStream );
    procedure WriteFontRecord( AStream: TStream; AFontInfo: TFontInfo );
    procedure WriteHeaderRecord( AStream: TStream );
    procedure WriteFooterRecord( AStream: TStream );
    procedure WriteFormatTable( AStream: TStream );
    procedure WriteFormatRecord( AStream: TStream; AFormat: string );
    procedure WriteWindowProtectRecord( AStream: TStream );
    procedure WriteFXTable( AStream: TStream );
    procedure WriteFXRecord( AStream: TStream; AFormatInfo: TFormatInfo );
    procedure WriteStyleTable( AStream: TStream );
    procedure WriteColumnInfoTable( AStream: TStream );
    procedure WriteColumnInfoRecord( AStream: TStream; AColumnInfo: TColumnInfo );
    procedure WriteCells( AStream: TStream );
    procedure WriteCellValue( AStream: TStream; ACellInfo: TCellInfo );

    procedure WriteCellValueEmpty( AStream: TStream; ACellInfo: TCellInfo );
    procedure WriteCellValueString( AStream: TStream; ACellInfo: TCellInfo );
    procedure WriteCellValueNumeric( AStream: TStream; ACellInfo: TCellInfo );
    procedure WriteCellValueDate( AStream: TStream; ACellInfo: TCellInfo );

    function HashCode( const ARow, AColumn: Integer ): Int64;
  public
    constructor Create;
    destructor Destroy; override;

    procedure SetColumnWidth( const AColumn, AWidth: Integer );
    function GetCellInfo( const ARow, AColumn: Integer ): TCellInfo;

    procedure SaveToFile( const AFilePath: string );
    procedure SaveToStream( AStream: TStream );

    property CodePage: Word read FCodePage write SetCodePage;
    property UserName: string read FUserName write SetUserName;
  end;

implementation

uses
  Variants;

function FontInfoComparer( const Left, Right: TFontInfo ): Integer;
begin
  Result := -1;
  if Left.Equals( Right ) then
    Result := 0;
end;

function FormatInfoComparer( const Left, Right: TFormatInfo ): Integer;
begin
  Result := -1;
  if Left.Equals( Right ) then
    Result := 0;
end;

function ColumnInfoComparer( const Left, Right: TColumnInfo ): Integer;
begin
  Result := -1;
  if Left.Equals( Right ) then
    Result := 0;
end;

{ TExcelDocument }

procedure TExcelDocument.BuildInternalTables;
var
  CellInfo: TCellInfo;
  FormatInfo: TFormatInfo;
begin
  for CellInfo in FCells.Values do
  begin
    FormatInfo := TFormatInfo.Create( CellInfo );
    if FFxs.IndexOf( FormatInfo ) = -1 then
      FFxs.Add( FormatInfo )
    else
      FormatInfo.Free;

    CellInfo.FXIndex := FFxs.IndexOf(FormatInfo) + 21;
  end;
end;

constructor TExcelDocument.Create;
begin
  FCells := TDictionary<Int64, TCellInfo>.Create;
  FFormats := TList<string>.Create;
  FFonts := TObjectList<TFontInfo>.Create( TFontInfoComparer.Construct(FontInfoComparer) );
  FFxs := TObjectList<TFormatInfo>.Create( TFormatInfoComparer.Construct(FormatInfoComparer) );
  FRows := TList<Integer>.Create;
  FColumns := TObjectList<TColumnInfo>.Create( TColumnInfoComparer.Construct(ColumnInfoComparer) );

  FCodePage := DefaultSystemCodePage;

  FFonts.Add( TFontInfo.Create('Arial', 10, COLOR_DEFAULT) );
  FFonts.Add( TFontInfo.Create('Arial', 10, COLOR_DEFAULT) );
  FFonts.Add( TFontInfo.Create('Arial', 10, COLOR_DEFAULT) );
  FFonts.Add( TFontInfo.Create('Arial', 10, COLOR_DEFAULT) );

  FFormats.Add('General');
  FFormats.Add('0');
  FFormats.Add('0.00');
  FFormats.Add('#,##0');
  FFormats.Add('#,##0.00');
  FFormats.Add('($#,##0_);($#,##0)');
  FFormats.Add('($#,##0_);[Red]($#,##0)');
  FFormats.Add('($#,##0.00_);($#,##0.00)');
  FFormats.Add('($#,##0.00_);[Red]($#,##0.00)');
  FFormats.Add('0%');
  FFormats.Add('0.00%');
  FFormats.Add('0.00E+00');
  FFormats.Add('# ?/?');
  FFormats.Add('# ??/??');
  FFormats.Add('m/d/yy');
  FFormats.Add('d-mmm-yy');
  FFormats.Add('d-mmm');
  FFormats.Add('mmm-yy');
  FFormats.Add('h:mm AM/PM');
  FFormats.Add('h:mm:ss AM/PM');
  FFormats.Add('h:mm');
  FFormats.Add('h:mm:ss');
  FFormats.Add('m/d/yy h:mm');
  FFormats.Add('(#,##0_);(#,##0)');
  FFormats.Add('(#,##0_);[Red](#,##0)');
  FFormats.Add('(#,##0.00_);(#,##0.00)');
  FFormats.Add('(#,##0.00_);[Red](#,##0.00)');
  FFormats.Add('_(* #,##0_);_(* (#,##0);_(* \''-\''_);_(@_)');
  FFormats.Add('_($* #,##0_);_($* (#,##0);_($* \''-\''_);_(@_)');
  FFormats.Add('_(* #,##0.00_);_(* (#,##0.00);_(* \''-\''??_);_(@_)');
  FFormats.Add('_($* #,##0.00_);_($* (#,##0.00);_($* \''-\''??_);_(@_)');
  FFormats.Add('mm:ss');
  FFormats.Add('[h]:mm:ss');
  FFormats.Add('mm:ss.0');
  FFormats.Add('##0.0E+0');
  FFormats.Add('@');
end;

destructor TExcelDocument.Destroy;
var
  CellInfo: TCellInfo;
begin
  for CellInfo in FCells.Values do
    CellInfo.Free;

  FCells.Free;
  FFormats.Free;
  FFonts.Free;
  FFxs.Free;
  FRows.Free;
  FColumns.Free;

  inherited;
end;

function TExcelDocument.GetCellInfo(const ARow, AColumn: Integer): TCellInfo;
var
  Key: Int64;
  CellInfo: TCellInfo;
begin
  Key := HashCode( ARow, AColumn );

  if FCells.TryGetValue( Key, CellInfo ) then
  begin
    Result := CellInfo;
  end
  else
  begin
    if FRows.IndexOf( ARow ) = -1 then
      FRows.Add( ARow );

    CellInfo := TCellInfo.Create( Self );
    CellInfo.Row := ARow;
    CellInfo.Column := AColumn;

    FCells.Add( Key, CellInfo );

    Result := CellInfo;
  end;
end;

function TExcelDocument.HashCode(const ARow, AColumn: Integer): Int64;
begin
  Result := (Int64(ARow) shl 32) + Int64(AColumn);
end;

procedure TExcelDocument.SaveToFile(const AFilePath: string);
var
  Stream: TMemoryStream;
begin
  Stream := TMemoryStream.Create;
  try
    SaveToStream( Stream );
    Stream.SaveToFile( AFilePath );
  finally
    Stream.Free;
  end;
end;

procedure TExcelDocument.SaveToStream(AStream: TStream);
begin
  AStream.Position := 0;
  AStream.Size := 0;

  BuildInternalTables;

  WriteWordArray( AStream, clBegin );
  WriteAuthorRecord( AStream );
  WriteCodepageRecord( AStream );
  WriteFontTable( AStream );
  WriteHeaderRecord( AStream );
  WriteFooterRecord( AStream );
  WriteFormatTable( AStream );
  WriteWindowProtectRecord( AStream );
  WriteFXTable( AStream );
  WriteStyleTable( AStream );
  WriteColumnInfoTable( AStream );
  WriteCells( AStream );
  WriteWordArray( AStream, clEnd );
end;

procedure TExcelDocument.SetCodePage(const Value: Word);
begin
  FCodePage := Value;
end;

procedure TExcelDocument.SetColumnWidth(const AColumn, AWidth: Integer);
var
  ColumnInfo: TColumnInfo;
  Index: Integer;
begin
  ColumnInfo := TColumnInfo.Create;

  ColumnInfo.ColumnIndex := AColumn;
  ColumnInfo.Width := AWidth;

  Index := FColumns.IndexOf( ColumnInfo );

  if Index = -1 then
  begin
    FColumns.Add( ColumnInfo );
  end
  else
  begin
    FColumns.Items[Index].Width := AWidth;
    ColumnInfo.Free;
  end;
end;

procedure TExcelDocument.SetUserName(const Value: string);
begin
  FUserName := Value;
end;

procedure TExcelDocument.WriteAuthorRecord(AStream: TStream);
var
  Data: array[0..1] of Word;
  TempUserName: AnsiString;
  PadSize: Integer;
  Temp: array of Byte;
begin
  Data[0] := $005c;
  Data[1] := 32;

  WriteWordArray( AStream, Data );

  TempUserName := AnsiString(FUserName);
  TempUserName := Copy(TempUserName, 1, 31 );
  { TODO :  MBCSΰ Ȯؾ ϳ??? }

  AStream.Write( PAnsiChar(TempUserName)^, Length(TempUserName) );
  PadSize := 32 - Length(TempUserName);
  SetLength( Temp, PadSize );
  WriteByteArray( AStream, Temp );
end;

procedure TExcelDocument.WriteByteArray(AStream: TStream;
  const AArray: array of Byte);
begin
  AStream.Write( AArray, Length(AArray) * sizeof(Byte) );
end;

procedure TExcelDocument.WriteCells(AStream: TStream);
var
  I: Integer;
  CellInfo: TCellInfo;
begin
  for I In FRows do
  begin
    for CellInfo in FCells.Values do
    begin
      if CellInfo.Row = I then
        WriteCellValue( AStream, CellInfo );
    end;
  end;
end;

procedure TExcelDocument.WriteCellValue(AStream: TStream; ACellInfo: TCellInfo);
begin
  if VarIsType( ACellInfo.Value, varNull ) then
    WriteCellValueEmpty( AStream, ACellInfo )
  else if VarIsType( ACellInfo.Value, varDate ) then
    WriteCellValueDate( AStream, ACellInfo )
  else if VarIsNumeric( ACellInfo.Value ) then
    WriteCellValueNumeric( AStream, ACellInfo )
  else if VarIsStr( ACellInfo.Value ) then
    WriteCellValueString( AStream, ACellInfo )
  else
  begin
  end;
end;

procedure TExcelDocument.WriteCellValueDate(AStream: TStream;
  ACellInfo: TCellInfo);
var
  Data: array [0..4] of Word;
  Value: Double;
begin
  Value := ACellInfo.Value;

  Data[0] := RECORD_NUMBER;
  Data[1] := 14;
  Data[2] := ACellInfo.Row;
  Data[3] := ACellInfo.Column;
  Data[4] := ACellInfo.FXIndex;

  WriteWordArray( AStream, Data );
  AStream.Write( Value, sizeof(Value) );
end;

procedure TExcelDocument.WriteCellValueEmpty(AStream: TStream;
  ACellInfo: TCellInfo);
var
  Data: array [0..4] of Word;
begin
  Data[0] := RECORD_EMPTY;
  Data[1] := 6;
  Data[2] := ACellInfo.Row;
  Data[3] := ACellInfo.Column;
  Data[4] := 15;
  WriteWordArray( AStream, Data );
end;

procedure TExcelDocument.WriteCellValueNumeric(AStream: TStream;
  ACellInfo: TCellInfo);
var
  Data: array [0..4] of Word;
  Value: Double;
begin
  Value := ACellInfo.Value;

  Data[0] := RECORD_NUMBER;
  Data[1] := 14;
  Data[2] := ACellInfo.Row;
  Data[3] := ACellInfo.Column;
  Data[4] := ACellInfo.FXIndex;

  WriteWordArray( AStream, Data );
  AStream.Write( Value, sizeof(Value) );
end;

procedure TExcelDocument.WriteCellValueString(AStream: TStream;
  ACellInfo: TCellInfo);
var
  Value: string;
  Temp: AnsiString;
  Data: array [0..5] of Word;
begin
  Value := ACellInfo.Value;
  Temp := AnsiString( Value );
  Temp := Copy( Temp, 1, 255 );

  Data[0] := RECORD_LABEL;
  Data[1] := 8 + Length(Temp);
  Data[2] := ACellInfo.Row;
  Data[3] := ACellInfo.Column;
  Data[4] := ACellInfo.FXIndex;
  Data[5] := Length(Temp);

  WriteWordArray( AStream, Data );
  AStream.Write( PAnsiChar(Temp)^, Length(Temp) );
end;

procedure TExcelDocument.WriteCodepageRecord(AStream: TStream);
var
  Data: array [0..2] of Word;
begin
  Data[0] := RECORD_CODEPAGE;
  Data[1] := $2;
  Data[2] := FCodePage;
  WriteWordArray( AStream, Data );
end;

procedure TExcelDocument.WriteColumnInfoRecord(AStream: TStream;
  AColumnInfo: TColumnInfo);
var
  Data: array [0..7] of Word;
begin
  Data[0] := RECORD_COMUMNINFO;
  Data[1] := 12;
  Data[2] := AColumnInfo.ColumnIndex;
  Data[3] := AColumnInfo.ColumnIndex;
  Data[4] := Trunc(AColumnInfo.Width * 256 / 7);
  Data[5] := 15;
  Data[6] := 0;
  Data[7] := 0;

  WriteWordArray( AStream, Data );
end;

procedure TExcelDocument.WriteColumnInfoTable(AStream: TStream);
var
  ColumnInfo: TColumnInfo;
begin
  for ColumnInfo in FColumns do
    WriteColumnInfoRecord( AStream, ColumnInfo );
end;

procedure TExcelDocument.WriteFontRecord(AStream: TStream;
  AFontInfo: TFontInfo);
var
  Data: array [0..4] of Word;
  FontName: AnsiString;
  FontNameLen: Byte;
  Flags: Word;
begin
  FontName := AnsiString( AFontInfo.FontName );
  FontName := Copy( FontName, 1, 254 );
  FontNameLen := Length( FontName );
  Flags := 0;

  if fsBold in AFontInfo.FFontStyle then
    Flags := Flags or 1;
  if fsItalic in AFontInfo.FFontStyle then
    Flags := Flags or 2;
  if fsUnderline in AFontInfo.FFontStyle then
    Flags := Flags or 4;
  if fsStrikeOut in AFontInfo.FFontStyle then
    Flags := Flags or 8;

  Data[0] := RECORD_FONT;
  Data[1] := 7 + FontNameLen;
  Data[2] := AFontInfo.FontSize * 20;
  Data[3] := Flags;
  Data[4] := AFontInfo.FontColor;

  WriteWordArray( AStream, Data );
  AStream.Write( FontNameLen, sizeof(Byte) );
  AStream.Write( PAnsiChar(FontName)^, FontNameLen );
end;

procedure TExcelDocument.WriteFontTable(AStream: TStream);
var
  FontInfo: TFontInfo;
begin
  for FontInfo in FFonts do
    WriteFontRecord( AStream, FontInfo );
end;

procedure TExcelDocument.WriteFooterRecord(AStream: TStream);
var
  Data: array [0..1] of Word;
begin
  Data[0] := RECORD_FOOTER;
  Data[1] := 0;

  WriteWordArray( AStream, Data );
end;

procedure TExcelDocument.WriteFormatRecord(AStream: TStream; AFormat: string);
var
  Data: array [0..1] of Word;
  Format: AnsiString;
  FormatLen: Byte;
  Temp: Byte;
begin
  Format := AnsiString( AFormat );
  Format := Copy( Format, 1, 254 );
  FormatLen := Length( Format );

  Data[0] := RECORD_FORMAT;
  Data[1] := 1 + FormatLen;

  WriteWordArray( AStream, Data );
  AStream.Write( FormatLen, sizeof(Byte) );
  AStream.Write( PAnsiChar(Format)^, Length(Format) );
end;

procedure TExcelDocument.WriteFormatTable(AStream: TStream);
var
  Format: string;
begin
  for Format in FFormats do
    WriteFormatRecord( AStream, Format );
end;

procedure TExcelDocument.WriteFXRecord(AStream: TStream;
  AFormatInfo: TFormatInfo);
var
  Data: array [0..1] of Word;
  Value: array [0..3] of Byte;
  Attr: Integer;
  Alignment, BackgroundArea: Word;
  Rest: array [0..3] of Word;
begin
  Data[0] := RECORD_EXTENDED;
  Data[1] := $00C;
  WriteWordArray( AStream, Data );

  Value[0] := AFormatInfo.FontIndex;
  Value[1] := AFormatInfo.FormatIndex;
  Value[2] := $01;

  Attr := 0;
  if AFormatInfo.FontIndex > 0 then
    Attr := Attr or $02;
  if AFormatInfo.Alignment <> xGeneral then
    Attr := Attr or $04;
  if AFormatInfo.BackColor <> COLOR_DEFAULT then
    Attr := Attr or $10;
  Attr := Attr shl 2;
  Value[3] := Attr;
  WriteByteArray( AStream, Value );

  Alignment := Ord(AFormatInfo.Alignment);
  BackgroundArea := 1;

  if AFormatInfo.BackColor <> COLOR_DEFAULT then
  begin
    BackgroundArea := (BackgroundArea and (not $07C0)) or (AFormatInfo.BackColor and ($07C0 shr 6)) shl 6;
    BackgroundArea := (BackgroundArea and (not $F800)) or (COLOR_WINDOWTEXT and ($F800 shr 11)) shl 11;
  end
  else
    BackgroundArea := $CE00;

  Rest[0] := Alignment;
  Rest[1] := BackgroundArea;
  Rest[2] := 0;
  Rest[3] := 0;
  WriteWordArray( AStream, Rest );
end;

procedure TExcelDocument.WriteHeaderRecord(AStream: TStream);
var
  Data: array [0..1] of Word;
begin
  Data[0] := RECORD_HEADER;
  Data[1] := 0;

  WriteWordArray( AStream, Data );
end;

procedure TExcelDocument.WriteStyleTable(AStream: TStream);
const
  Data1: array [0..3] of Byte = ($10,$80,$03,$FF);
  Data2: array [0..11] of Byte = ($11,$00,$09,$43,$6F,$6D,$6D,$61,$20,$5B,$30,$5D);
  Data3: array [0..3] of Byte = ($12,$80,$04,$FF);
  Data4: array [0..14] of Byte = ($13,$00,$0C,$43,$75,$72,$72,$65,$6E,$63,$79,$20,$5B,$30,$5D);
  Data5: array [0..3] of Byte = ($00,$80,$00,$FF);
  Data6: array [0..3] of Byte = ($14,$80,$05,$FF);
var
  Header: array [0..1] of Word;

  procedure WriteData( Data: array of Byte );
  begin
    Header[1] := Length(Data);
    WriteWordArray( AStream, Header );
    WriteByteArray( AStream, Data );
  end;
begin
  Header[0] := RECORD_STYLE;

  WriteData( Data1 );
  WriteData( Data2 );
  WriteData( Data3 );
  WriteData( Data4 );
  WriteData( Data5 );
  WriteData( Data6 );
end;

procedure TExcelDocument.WriteWindowProtectRecord(AStream: TStream);
var
  Data: array [0..1] of Word;
begin
  Data[0] := RECORD_WINDOWPROTECT;
  Data[1] := 0;

  WriteWordArray( AStream, Data );
end;

procedure TExcelDocument.WriteWordArray(AStream: TStream;
  const AArray: array of Word);
begin
  AStream.Write( AArray, Length(AArray) * sizeof(Word) );
end;

procedure TExcelDocument.WriteFXTable(AStream: TStream);
const
  Data: array [0..167] of Word = (
  $0243, $00C, $0000, $03F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0001, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0001, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0002, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0002, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $F7F5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0000, $0001, $0000, $CE00, $0000, $0000,
  $0243, $00C, $2101, $FBF5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $1F01, $FBF5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $2001, $FBF5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $1E01, $FBF5, $FFF0, $CE00, $0000, $0000,
  $0243, $00C, $0901, $FBF5, $FFF0, $CE00, $0000, $0000 );
var
  FormatInfo: TFormatInfo;
begin
  WriteWordArray( AStream, Data );

  for FormatInfo in FFxs do
    WriteFXRecord( AStream, FormatInfo );
end;

{ TColumnInfo }

function TColumnInfo.Equals(Obj: TObject): Boolean;
begin
  Result := False;

  if not Assigned(Obj) then
    Exit;

  if Obj is TColumnInfo then
  begin
    Result := ColumnIndex = (Obj as TColumnInfo).ColumnIndex;
  end
  else
  begin
    Result := inherited;
  end;
end;

procedure TColumnInfo.SetColumnIndex(const Value: Integer);
begin
  FColumnIndex := Value;
end;

procedure TColumnInfo.SetWidth(const Value: Integer);
begin
  FWidth := Value;
end;

{ TFontInfo }

constructor TFontInfo.Create;
begin
  FFontColor := COLOR_DEFAULT;
end;

procedure TFontInfo.Assign(AFontInfo: TFontInfo);
begin
  FFontName := AFontInfo.FontName;
  FFontSize := AFontInfo.FontSize;
  FFontStyle := AFontInfo.FontStyle;
  FFontColor := AFontInfo.FontColor;
end;

constructor TFontInfo.Create(const AFontName: string; const AFontSize: Integer;
  const AFontColor: Word);
begin
  Create;

  FFontName := AFontName;
  FFontSize := AFontSize;
  FFontColor := AFontColor;
end;

function TFontInfo.Equals(Obj: TObject): Boolean;
var
  FontInfo: TFontInfo;
begin
  Result := False;

  if Obj is TFontInfo then
  begin
    FontInfo := Obj as TFontInfo;
    if (FFontSize = FontInfo.FontSize) and
      (FFontName = FontInfo.FontName) and
      (FFontStyle = FontInfo.FontStyle) and
      (FFontColor = FontInfo.FontColor) then
      Result := True;
  end;
end;

procedure TFontInfo.SetFontColor(const Value: Word);
begin
  FFontColor := Value;
end;

procedure TFontInfo.SetFontName(const Value: String);
begin
  FFontName := Value;
end;

procedure TFontInfo.SetFontSize(const Value: Integer);
begin
  FFontSize := Value;
end;

procedure TFontInfo.SetFontStyle(const Value: TFontStyles);
begin
  FFontStyle := Value;
end;

{ TCellInfo }

constructor TCellInfo.Create( AExcelDocument: TExcelDocument );
begin
  FDocument := AExcelDocument;

  FForeColor := COLOR_DEFAULT;
  FBackColor := COLOR_DEFAULT;
  FFont := TFontInfo.Create( 'Arial', 10, COLOR_DEFAULT );
end;

destructor TCellInfo.Destroy;
begin
  FFont.Free;

  inherited;
end;

procedure TCellInfo.SetAlignment(const Value: TAlignment);
begin
  FAlignment := Value;
end;

procedure TCellInfo.SetBackColor(const Value: Word);
begin
  FBackColor := Value;
end;

procedure TCellInfo.SetColumn(const Value: Integer);
begin
  FColumn := Value;
end;

procedure TCellInfo.SetDocument(const Value: TExcelDocument);
begin
  FDocument := Value;
end;

procedure TCellInfo.SetFont(const Value: TFontInfo);
begin
  FFont.Assign( Value );
end;

procedure TCellInfo.SetForeColor(const Value: Word);
begin
  FForeColor := Value;
end;

procedure TCellInfo.SetFormat(const Value: string);
begin
  FFormat := Value;

  if FDocument.FFormats.IndexOf( Value ) = -1 then
    FDocument.FFormats.Add( Value );
end;

procedure TCellInfo.SetFXIndex(const Value: Byte);
begin
  FFXIndex := Value;
end;

procedure TCellInfo.SetRow(const Value: Integer);
begin
  FRow := Value;
end;

procedure TCellInfo.SetValue(const Value: Variant);
begin
  FValue := Value;

  if VarIsType( Value, varDate ) then
  begin
    FFormat := FDocument.FFormats[15];
  end;
end;

{ TCell }

constructor TCell.Create(const ARow, AColumn: Integer;
  ADocument: TExcelDocument);
begin
  FDocument := ADocument;
  FCellInfo := FDocument.GetCellInfo( ARow, AColumn );
end;

function TCell.GetAlignment: TAlignment;
begin
  Result := FCellInfo.Alignment;
end;

function TCell.GetBackColor: Word;
begin
  Result := FCellInfo.BackColor;
end;

function TCell.GetColumn: Integer;
begin
  Result := FCellInfo.Column;
end;

function TCell.GetFont: TFontInfo;
begin
  Result := FCellInfo.Font;
end;

function TCell.GetForeColor: Word;
begin
  Result := FCellInfo.ForeColor;
end;

function TCell.GetFormat: string;
begin
  Result := FCellInfo.Format;
end;

function TCell.GetRow: Integer;
begin
  Result := FCellInfo.Row;
end;

function TCell.GetValue: Variant;
begin
  Result := FCellInfo.Value;
end;

procedure TCell.SetAlignment(const Value: TAlignment);
begin
  FCellInfo.Alignment := Value;
end;

procedure TCell.SetBackColor(const Value: Word);
begin
  FCellInfo.BackColor := Value;
end;

procedure TCell.SetFont(const Value: TFontInfo);
begin
  FCellInfo.Font := Value;
end;

procedure TCell.SetForeColor(const Value: Word);
begin
  FCellInfo.ForeColor := Value;
end;

procedure TCell.SetFormat(const Value: string);
begin
  FCellInfo.Format := Value;

  if FDocument.FFormats.IndexOf( Value ) = -1 then
    FDocument.FFormats.Add( Value );
end;

procedure TCell.SetValue(const Value: Variant);
begin
  FCellInfo.Value := Value;
end;

{ TFormatInfo }

constructor TFormatInfo.Create(ACellInfo: TCellInfo);
var
  FontInfo: TFontInfo;
begin
  FCellInfo := ACellInfo;

  if FCellInfo.Format = '' then
    FFormatIndex := 0
  else
    FFormatIndex := FCellInfo.Document.FFormats.IndexOf( FCellInfo.Format );

  FontInfo := TFontInfo.Create;
  FontInfo.Assign( FCellInfo.Font );

  FFontIndex := FCellInfo.Document.FFonts.IndexOf( FontInfo );
  if FFontIndex = -1 then
  begin
    FCellInfo.Document.FFonts.Add( FontInfo );
    FFontIndex := FCellInfo.Document.FFonts.IndexOf( FontInfo );
  end
  else
    FontInfo.Free;

  if FFontIndex > 3 then
    Inc( FFontIndex );
end;

function TFormatInfo.Equals(Obj: TObject): Boolean;
var
  FormatInfo: TFormatInfo;
begin
  Result := False;

  if Obj is TFormatInfo then
  begin
    FormatInfo := Obj as TFormatInfo;
    Result := (FFontIndex = FormatInfo.FontIndex) and
      (FFormatIndex = FormatInfo.FormatIndex) and
      (ForeColor = FormatInfo.ForeColor) and
      (BackColor = FormatInfo.BackColor) and
      (Alignment = FormatInfo.Alignment);
  end;
end;

function TFormatInfo.GetAlignment: TAlignment;
begin
  Result := FCellInfo.Alignment;
end;

function TFormatInfo.GetBackColor: Word;
begin
  Result := FCellInfo.BackColor;
end;

function TFormatInfo.GetFont: TFontInfo;
begin
  Result := FCellInfo.Font;
end;

function TFormatInfo.GetForeColor: Word;
begin
  Result := FCellInfo.ForeColor;
end;

function TFormatInfo.GetFormat: string;
begin
  Result := FCellInfo.Format;
end;

end.
